/*
 ============================================================================
 Project    : Parallel Genetic Algorithm for the SUPBR Problem
                (version with multiple populations)
 File       : main.c
 Author     : Lucas Angelo
                Group of Theory of Computation
                Universidade de Brasilia (UnB) - Brazil
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>//time
#include <mpi.h>
#include <math.h>
#include <string.h>

#include "operadores.h"
#include "structs_ga.h"
#include "calc_fitness.h"
#include "GerarPopuInicial.c"


#define WORKTAG 1
#define DIETAG 2
#define EVALTAG 3
#define GOTAG 4

#define DEBUG 0
#define TAM_BUFFER 1000

void leer_permutacion(permutacion *perm);
void slave(parameters *params);

void leer_permutacion(permutacion *perm){
	int i,n;
	scanf("%d",&n);//leer tamanho de permutacion
	perm->tamanho = n;
	for(i=0; i<n; i++){
		scanf("%d",&perm->pi[i]);
	}
}


//flag em uma roda vai comecar no inicio outro no final
int Imigracao( int imigracao, int sizePopulation, int fitSemelhante, poblacion *pob, int flag){
    int i,indice;
    int porcento= sizePopulation *0.2;
    switch (imigracao)
    {
        case 1: // pior
            indice= (sizePopulation - porcento) +(rand() % porcento);
            break;
        case 2: //aleatorio
            indice=(rand() % sizePopulation);
            break;
        case 3://semelhante
            indice=(rand() % sizePopulation);
            if (flag)
            {
                for (i=0; i<sizePopulation; i++)
                {
                    if (pob->cromosomas[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end if
            else
            {
                for (i=(sizePopulation-1); i>=0; i--)
                {
                    if (pob->cromosomas[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end else
            break;
    }//end switch
    


     if (indice > (sizePopulation-1) )
     {
         printf("ERRO AQUI NO IMIGRACAO\n\n");
     }


    return indice;
    
    
}// end function Imigracao


//busca os indicies a ser enviado para outra ilha
int * buscaIndex(int qtdIndividual, int typeIndividual, int sizePopulation)
{
    int i,j;
    int * indices= malloc (qtdIndividual * sizeof (int));
    if (indices==NULL)
        printf("ATENCAO PERCA DE MEMORIA, GA_Paralle_pop.c, LINE 236\n");

    switch (typeIndividual) {
        case 1://melhor
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=i;
            }
            break;
            
        case 2: // pior
            i=0;
            j=(sizePopulation-1);
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=j;
                j--;
                
            }//end for
            break;
            
        case 3: //aleatorio
            for (i=0; i<qtdIndividual; i++)
            {
                
                indices[i]=(rand() % sizePopulation);
                if (indices[i]>= sizePopulation)
                {
                    printf("Erro aqui no buscaIndex\n\n");
                }
            }
            break;
    }//end switch
    
    
    return indices;
    
}// end buscaIndex




void sendALL( MPI_Comm communicator, int world_rank, poblacion *pob)
{
    
    int world_size, j, iii,i;
    MPI_Comm_size(communicator, &world_size);
    permutacion enviar,recebe;
    
    // If we are the root process, send our best data to everyone
    enviar.tamanho=pob->cromosomas[0].tamanho;
    enviar.fitness=pob->cromosomas[0].fitness;
    for (i=0; i<pob->cromosomas[0].tamanho; i++)
        enviar.pi[i]=pob->cromosomas[0].genes[i];
    
    
    
    if (world_rank!=0)
        
    {
        int error_code= MPI_Send(&enviar,//the data
                                 sizeof(permutacion),
                                 MPI_BYTE,
                                 0,//destination process
                                 0,//tag
                                 communicator);
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
            
        }
        
    }
    
    else
    {
        for( iii = 0; iii < world_size; iii++)
        {
            
            
            if (iii != world_rank)
            {
                int error_code=MPI_Recv(&recebe,
                                        sizeof(permutacion),
                                        MPI_BYTE,
                                        iii, // processo que enviou
                                        0,//tag
                                        communicator,
                                        MPI_STATUS_IGNORE);
                
                if (error_code != MPI_SUCCESS)
                {
                    char error_string[BUFSIZ];
                    int length_of_error_string;
                    MPI_Error_string(error_code, error_string, &length_of_error_string);
                    printf("%3d: %s\n", world_rank, error_string);
                }//end if
                
                int controle=0;
                for (j=0; j<pob->tamanhoPoblacion; j++)
                {
                    if (recebe.fitness < pob->cromosomas[j].fitness)
                    {
                        pob->cromosomas[j].fitness=recebe.fitness;
                        for (i=0; i<pob->cromosomas[j].tamanho; i++)
                        {
                            pob->cromosomas[j].genes[i]=recebe.pi[i];
                        }//end for
                        controle=1;
                    }//end if
                    if (controle)
                        break;
                }//end for
            }//end if
        }//end for
        
    }//else
    
}// end function








void my_bcast( MPI_Comm communicator, int world_rank, poblacion *pob, int emigracao, int qtdIndividual, int typeIndividual, int imigracao) {

    int world_size, j, iii,i,jj,index;

    MPI_Comm_size(communicator, &world_size);

    permutacion enviar,recebe;

    int flag=0;

    int * IndicesEnvia=buscaIndex(qtdIndividual,typeIndividual,pob->tamanhoPoblacion);

    for (jj=0; jj< qtdIndividual; jj++)

    {

        

        // If we are the root process, send our best data to everyone

        enviar.tamanho=pob->cromosomas[IndicesEnvia[jj]].tamanho;

        enviar.fitness=pob->cromosomas[IndicesEnvia[jj]].fitness;

        for (i=0; i<pob->cromosomas[IndicesEnvia[jj]].tamanho; i++)

            enviar.pi[i]=pob->cromosomas[IndicesEnvia[jj]].genes[i];

        

        

        

        if (world_rank!=0)

            

        {

            int error_code= MPI_Send(&enviar,//the data

                                     sizeof(permutacion),

                                     MPI_BYTE,

                                     0,//destination process

                                     0,//tag

                                     communicator);

            if (error_code != MPI_SUCCESS)

            {

                char error_string[BUFSIZ];

                int length_of_error_string;

                MPI_Error_string(error_code, error_string, &length_of_error_string);

                printf("%3d: %s\n", world_rank, error_string);

                

            }

            

        }

        

        else

        {

            for( iii = 0; iii < world_size; iii++)

            {

                

                

                if (iii != world_rank)

                {

                    int error_code=MPI_Recv(&recebe,

                                            sizeof(permutacion),

                                            MPI_BYTE,

                                            iii, // processo que enviou

                                            0,//tag

                                            communicator,

                                            MPI_STATUS_IGNORE);

                    

                    if (error_code != MPI_SUCCESS)

                    {

                        char error_string[BUFSIZ];

                        int length_of_error_string;

                        MPI_Error_string(error_code, error_string, &length_of_error_string);

                        printf("%3d: %s\n", world_rank, error_string);

                    }//end if

                    

                    if (flag) {

                        flag=0;

                    }

                    else

                        flag=1;

                    

                    index=Imigracao( imigracao, pob->tamanhoPoblacion, recebe.fitness,  pob,  flag);

                    //printf("index %d tamanho total %d\n", index,pob->sizePopulation);

                    if (index>=pob->tamanhoPoblacion) {

                        index=pob->tamanhoPoblacion-1;

                        printf("DEU ERRADO NA PARTE DE IMIGRACAO \n");

                    }

                    

                    pob->cromosomas[index].fitness=recebe.fitness;

                    

                    for (j=0; j<pob->cromosomas[index].tamanho; j++)

                        pob->cromosomas[index].genes[j]=recebe.pi[j];

                }//end if

            }//end for

            

        }//else

        

    }//end for

    

    

    free(IndicesEnvia);

    

}//end function broadcast








//implementa topologia anel
//flag é pra marcar se for igual a 1 significa q esta terminando as geracoes e todos as ilhas necessitas saber os resultados
void my_bcastanel(MPI_Comm communicator, int world_rank, poblacion *pob, int emigracao, int qtdIndividual, int typeIndividual, int imigracao) {
    int world_size, j,i,ii,error_code, index;
    int envio=-1;
    int flag=0;
    int receptor=-1;
    int * IndicesEnvia;
    MPI_Comm_size(communicator, &world_size);
    permutacion enviar,recebe;
    
    //buscar indices para ser enviado
    
    IndicesEnvia=buscaIndex(qtdIndividual,typeIndividual,pob->tamanhoPoblacion);
    
    for (i=0; i< qtdIndividual; i++)
    {
        // If we are the root process, send our data to everyone
        enviar.tamanho=pob->cromosomas[IndicesEnvia[i]].tamanho;
        enviar.fitness=pob->cromosomas[IndicesEnvia[i]].fitness;
        for (j=0; j<pob->cromosomas[IndicesEnvia[i]].tamanho; j++)
            enviar.pi[j]=pob->cromosomas[IndicesEnvia[i]].genes[j];
        
        //sent to world_rank+1
        
        if ((world_rank+1)< world_size)
        {
            envio=(world_rank+1);
            
            if (world_rank==0)
                receptor=(world_size-1);
            
            else
                receptor=(world_rank-1);
            
            
        }//end if
        
        else
        {
            envio=0;
            receptor=(world_rank-1);
            
        }
        
        
        /***************Parte onde trata emigracao segundo exclusao ************************/
        if (emigracao==2) {
            pob->cromosomas[IndicesEnvia[i]].fitness=pob->cromosomas[IndicesEnvia[i]].tamanho;
        }
        
        ///sent message
        error_code= MPI_Send(&enviar,//the data
                             sizeof(permutacion),
                             MPI_BYTE,
                             envio,//destination process
                             0,//tag
                             communicator);
        
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
            
        }
        
        
        
        
        error_code=MPI_Recv(&recebe,
                            sizeof(permutacion),
                            MPI_BYTE,
                            receptor, // processo que enviou
                            0,//tag
                            communicator,
                            MPI_STATUS_IGNORE);
        
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
        }//end if
        
        
        
        
            if (flag) {
                flag=0;
            }
            else
                flag=1;
            
            index=Imigracao( imigracao, pob->tamanhoPoblacion, recebe.fitness,  pob,  flag);
            //printf("index %d tamanho total %d\n", index,pob->sizePopulation);
            if (index>=pob->tamanhoPoblacion) {
                index=pob->tamanhoPoblacion-1;
                printf("DEU ERRADO NA PARTE DE IMIGRACAO \n");
            }
            
            pob->cromosomas[index].fitness=recebe.fitness;
            
            for (ii=0; ii<pob->cromosomas[index].tamanho; ii++)
                pob->cromosomas[index].genes[ii]=recebe.pi[ii];
            
        
        
    }//end for
    
    free(IndicesEnvia);
    
    
    
}//end function anel


void mostrar_fitness(poblacion *pob, int tamanho){
    
    int i;
    //imprimir poblacion actual
    for(i=pob->basePoblacion; i<pob->topePoblacion; i++){
        printf("(%d): ",i);
        
            printf("%d) %d ", i, pob->cromosomas[i].fitness);
        printf("\n");
    }
    
}




void GA(MPI_Comm communicator, int world_rank,parameters *params, float prob_crossover,float porcentaje_selec,float porcentaje_reemp,float prob_mutacion, int emigracao, int typeIndividual, int qtdIndividual,int imigracao, float intervaloMigracao)
{
    permutacion perm;
    poblacion pob, pob2;
    int mejor_solucion=-1;//mejor solucion encontrada
    int i, num_generaciones,nro_generacion, world_size;
    MPI_Comm_size(communicator, &world_size);
    
    //Read Permutation
    leer_permutacion_de_archivo("i", &perm);
    params->size_population = -1;
    params->number_generations = -1;
    params->num_eval_fit_function = -1;
    params->number_reversals = -1;
    
    //processo 0 gera a populacao
    if (world_rank==0) {
        poblacion_inicial2(&pob2,&perm,porcentaje_reemp, porcentaje_selec, world_size, params);//crear poblacion inicial
        construirPopInit(world_size,&pob2, perm.tamanho, &perm);
        for(i=0;i<pob2.tamanhoTotal;i++){
            free(pob2.cromosomas[i].genes);
        }
        free(pob2.cromosomas);
    }
    MPI_Barrier(MPI_COMM_WORLD);
    
    //aqui cada processo carrega tua populacao
    CarregarPopInitial(world_rank, porcentaje_reemp, porcentaje_selec, &pob,&perm);
    int num_eval_fit_function = 0;
    fitness(&pob,1,&num_eval_fit_function);//calcular fitness pob inicial
    //mostrar_fitness(&pob, perm.tamanho);
    mejor_solucion = pob.cromosomas[0].fitness;
    
    
    num_generaciones = perm.tamanho;//num de generaciones = a tamanho de permutacion
    int contador = num_generaciones * intervaloMigracao;
    for(nro_generacion=2; nro_generacion<=num_generaciones; nro_generacion++)
    {
        
        seleccion(&pob,nro_generacion,&mejor_solucion);//seleccion:ordenar poblacion
        crossover3(&pob, &perm, prob_crossover);//crossover:generar descendencia
        mutacion(&pob, &perm, prob_mutacion);//mutacion:actuar sobre toda la descendencia (de tam_desc hasta tamanho de descencia)
        fitness(&pob,nro_generacion,&num_eval_fit_function);//calcular fitness de descendencia
        reemplazo(&pob);//reemplazar descendencia en nueva generacion
        
        contador--;
        if (contador==0)
        {
           my_bcast(communicator, world_rank, &pob, emigracao, qtdIndividual,typeIndividual, imigracao);
            
            contador = num_generaciones * intervaloMigracao;
        }
        
        if (nro_generacion==num_generaciones)
        {
            sendALL(communicator, world_rank, &pob);
            
        }//end if
        
        MPI_Barrier(MPI_COMM_WORLD);
        num_eval_fit_function=0;
    }
    
    
    if (world_rank==0)
    {    seleccion(&pob,nro_generacion,&mejor_solucion);//seleccion:ordenar poblacion
         printf("CP GA: %d\n",mejor_solucion);
    }
    
    for(i=0;i<pob.tamanhoTotal;i++){
        free(pob.cromosomas[i].genes);
    }
    free(pob.cromosomas);
    
    
}

int main(int argc,char **argv)
{
    int seed[24] = {408302,680502,414942,829352,627420, 753780, 110291,444847, 127844, 525817,887214,115510,112376,510836,310830,390833,560813,260810,908811,414819,939819,287817,364410,194419};
    unsigned int seed2 = time(NULL);
    float prob_crossover=		0.9;
    float porcentaje_selec=		0.9;
    float porcentaje_reemp=		0.3;
    float prob_mutation=		0.01;
    int emigracao=			1;
    int typeIndividual=			3;
    int qtdIndividual=			1;
    int imigracao=			2;
    float intervaloMigracao= 		0.1;
    int world_rank;
    MPI_Init(NULL, NULL);
    parameters params;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    srand (seed2);//initialize random generator
    params.seed =seed[world_rank]+seed2;
    GA(MPI_COMM_WORLD, world_rank, &params, prob_crossover,porcentaje_selec,porcentaje_reemp, prob_mutation,emigracao,typeIndividual,qtdIndividual,imigracao,intervaloMigracao);
    MPI_Finalize();
    return 0;
}






